-- ********************************************************************************
-- Binarizer for HEVC h.265       
-- 01/03/2017                 
-- Norwegian University of Science and Technology              
-- Lars Erik Songe Paulsen             
-- ********************************************************************************

-- ********************************************************************************
-- TODO LIST:    
-- ********************************************************************************
-- variable K  
-- Limit coeff_abs_level_greater1(ALG1) to 8 bits output.                 
-- ********************************************************************************

--pragma synthesis_off
  -- your simulation-only code
--pragma synthesis_on

library ieee;
library std;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all; -- http://www1.pldworld.com/@xilinx/html/technote/tool/manual/15i_doc/fndtn/vhd/vhd10_3.htm
use std.textio.all;
use ieee.math_real.all;

library work;
use work.BinarizerParameters.all;

entity Binarizer is

    port 	
    (
    Clk                         : in std_logic;
    DataIn                      : in std_logic_vector((16*CoeffWidth)-1 downto 0);
    ScanDir                     : in std_logic_vector(1 downto 0);
    Resetn                      : in std_logic;
    StartBinarizer              : in std_logic;
    DataOut                     : out std_logic_vector(OutputWidth-1 downto 0); 
    DataLength                  : out std_logic_vector(OutputWidthLength-1 downto 0);
    PrefixLength                : out std_logic_vector(OutputWidthLength-1 downto 0);
    Finished                    : out std_logic
    );

end Binarizer;

architecture struct of Binarizer is

    -- ---------------------------------------------------------------------------
    -- Type declarations
    -- ---------------------------------------------------------------------------

    type N_low   is array(0 to 18)                     of std_logic_vector(15 downto 0);
    type k_table is array(0 to  4)                     of N_low;

    type BinarizeStateType is(
                    read_input, 
                    write_last_sig_coeff_x_prefix, 
                    write_last_sig_coeff_y_prefix, 
                    write_sig_coeff_flag,
                    write_coeff_abs_level_greater1,
                    write_coeff_abs_level_greater2,
                    write_coeff_abs_level_remaining,
                    write_coeff_sign_flag,
                    write_finished
                );

    -- ---------------------------------------------------------------------------
    -- Functions
    -- ---------------------------------------------------------------------------
    function string_to_binary(inp: string) return std_logic_vector is 
        variable temp: std_logic_vector(inp'length-1 downto 0) := (others => 'X'); 
    begin 
        for i in inp'range loop 
            case inp(i) is 
                when '0'    => temp(i-1) := '0'; 
                when '1'    => temp(i-1) := '1'; 
                when others => temp(i-1) := 'X'; 
            end case; 
        end loop;
        return temp; 
    end function string_to_binary;

    impure function Initk_table (RomFileName : in string) return k_table is
        FILE romfile : text is in RomFileName;
        variable RomFileLine : line;
        variable rom : k_table;
        variable TestString : string(16 downto 1);
    begin
        for i in 0 to 4 loop
            for j in 0 to 18 loop
                readline(romfile, RomFileLine);
                read(RomFileLine, TestString);
                rom(i)(j) := string_to_binary(TestString)(15 downto 0);
            end loop;
        end loop;
        return rom;
    end function;

    -- ---------------------------------------------------------------------------
    -- Signal declarations
    -- ---------------------------------------------------------------------------

    signal kTable         : k_table :=  Initk_table("Binarizer_Tables\k_table.txt");


    type transform_block is array(15 downto 0) of integer range -(2**(coeffwidth-1)) to (2**(coeffwidth-1));

    signal BinarizeState : BinarizeStateType;
	
    signal coefficients : Transform_Block;
    signal coeff_abs_level_greater2_flag : std_logic_vector(0 to 15);
    signal ABS_index : integer range 0 to 15;
    signal ABS_started : std_logic;
    signal ABS_done : std_logic;
	signal ABS_level_writeout : integer range 0 to 32768;
    signal k : integer range 0 to 4;

	
begin

    Next_ABS : process(Clk, DataIn, Resetn, StartBinarizer)
        variable ABS_index_cycler : integer range 0 to 15;
		
    begin
        if Resetn = '0' then
			ABS_index_cycler := 0;
        elsif rising_edge(Clk) then
            case BinarizeState is			                    
                when write_coeff_abs_level_greater1 | write_coeff_abs_level_remaining =>
                    find_next_ALG2_loop : for ABS_index_cycler in 0 to 15 loop
                        if (ABS_index_cycler <= ABS_index) then
                        --spin TODO FIX end condition
                        else
                            if(coeff_abs_level_greater2_flag(ABS_index_cycler) = '1') then
                                report "ALG found at ABS_index[" & integer'image(ABS_index_cycler) & "]";
                                ABS_index <= ABS_index_cycler;
								ABS_level_writeout <= coefficients(ABS_index_cycler);
                                if(ABS_index_cycler = 15) then
                                    ABS_done <= '1';
                                else
                                    exit find_next_ALG2_loop;
                                end if;
                            end if;
                        end if;
                    end loop;
                when others =>
                    ABS_index <= 0;
                    ABS_done  <= '0';
            end case;
        end if;
    end process;

    Binarize : process(Clk, DataIn, Resetn, StartBinarizer)
    
    -- ---------------------------------------------------------------------------
    -- Variable declarations
    -- ---------------------------------------------------------------------------
    
    variable sig_coeff_flag : std_logic_vector(0 to 15);
    variable sig_coeff_flag_found : std_logic;
    variable sig_coeff_flag_index : integer range OutputWidth-1 downto OutputWidth-14;
    variable coeff_sign_flag : std_logic_vector(0 to 15);
    variable ALG1_index  : integer range 0 to 8;   
    variable SIGN_index  : integer range 0 to 15;   
    variable Scan_Direction : Scan_Directions := DiagonalScan; -- TODO: Add all scan directions
	variable coeff_abs_level_greater1_flag : std_logic_vector(0 to 15);
    variable result : std_logic_vector(17 downto 0);
    --variable xorresult : std_logic_vector(16 downto 0);
    variable xorresult : integer range 0 to 17;

    begin
        if Resetn = '0' then
            sig_coeff_flag := (others => '0');
            sig_coeff_flag_found := '0';
            sig_coeff_flag_index := OutputWidth-1;
            ALG1_index     := 0;
            SIGN_index     := 0;
            DataOut        <= (others => '0');
            DataLength     <= (others => '0');
            BinarizeState  <= read_input;
            Finished       <= '1';
            coeff_abs_level_greater1_flag := (others => '0');
            coeff_abs_level_greater2_flag <= (others => '0');
            k   <= 0;
            PrefixLength <= (others => '0');
			
        elsif rising_edge(Clk) then
            case BinarizeState is
                when read_input =>
                    BinarizeState <= write_last_sig_coeff_x_prefix;
                    Finished      <= '0';
					
					for i in 0 to 15 loop
						-- ---------------------------------------------------------------------------
						-- Read absolute values.
						-- ---------------------------------------------------------------------------	
						coefficients(Scan_Direction(i))  <= to_integer(abs(signed(DataIn(((CoeffWidth*(i+1))-1)   downto  (CoeffWidth*i)))));
						
						-- ---------------------------------------------------------------------------
						-- Read sign.
						-- ---------------------------------------------------------------------------
						coeff_sign_flag(Scan_Direction(i))   := DataIn((CoeffWidth*(i+1))-1);
						
						-- ---------------------------------------------------------------------------
						-- Find non-zero data.
						-- ---------------------------------------------------------------------------	
						sig_coeff_flag(Scan_Direction(i))  := or_reduce(DataIn(((CoeffWidth*(i+1))-2)   downto  (CoeffWidth*i))); 
					
						-- ---------------------------------------------------------------------------
						-- Find >1 data and >2 data.
						-- ---------------------------------------------------------------------------	
						case DataIn((CoeffWidth*(i+1))-1) is
							when '0' =>
								coeff_abs_level_greater1_flag(Scan_Direction(i))  	:= or_reduce(DataIn(((CoeffWidth*(i+1))-2)  downto  ((CoeffWidth*i)+1)));
								coeff_abs_level_greater2_flag(Scan_Direction(i))  	<= (DataIn(CoeffWidth*i)   and DataIn((CoeffWidth*i)+1)) 
																					or or_reduce(DataIn(((CoeffWidth*(i+1))-2) downto ((CoeffWidth*i)+2))); 
							when '1' =>
								coeff_abs_level_greater1_flag(Scan_Direction(i))  	:= nand_reduce(DataIn(((CoeffWidth*(i+1))-2)  downto  (CoeffWidth*i)));
								coeff_abs_level_greater2_flag(Scan_Direction(i))  	<= (DataIn(CoeffWidth*i)   and (not(DataIn((CoeffWidth*i)+1)))) 
																					or nor_reduce(DataIn(((CoeffWidth*(i+1))-2)  downto((CoeffWidth*i)+2))); 
							when others =>
						end case;
					end loop;

                when write_last_sig_coeff_x_prefix =>
                    BinarizeState <= write_last_sig_coeff_y_prefix;
                    write_last_sig_coeff_x_prefix_loop : for index in 0 to 15 loop
                        if(sig_coeff_flag(index) = '1') then
                            case index is                                
                                when 0  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 1  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 2  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 3  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 4  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 5  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength)); 
                                when 6  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 7  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 8  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 9  =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 10 =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 11 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 12 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 13 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 14 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 15 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                            end case;
                            exit write_last_sig_coeff_x_prefix_loop;
                        end if;
                    end loop;
                when write_last_sig_coeff_y_prefix =>
                    BinarizeState <= write_sig_coeff_flag;
                    write_last_sig_coeff_y_prefix_loop : for index in 0 to 15 loop
                        if(sig_coeff_flag(index) = '1') then
                            case index is                                
                                when 0  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 1  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 2  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 3  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength)); 
                                when 4  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 5  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 6  =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength)); 
                                when 7  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 8  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 9  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 10 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 11 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 12 =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 13 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 14 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 15 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                            end case;
                            exit write_last_sig_coeff_y_prefix_loop;
                        end if;
                    end loop;
                when write_sig_coeff_flag =>
					BinarizeState <= write_coeff_abs_level_greater1;
                    write_sig_coeff_flag_loop : for index in 0 to 15 loop
                        if sig_coeff_flag_found = '1' then
                            DataOut(sig_coeff_flag_index) <= sig_coeff_flag(index);
                            sig_coeff_flag_index := sig_coeff_flag_index - 1; 
                       elsif(sig_coeff_flag(index) = '1') then
                            sig_coeff_flag_found := '1';
                            DataLength <= std_logic_vector(to_unsigned(15-index,OutputWidthLength));                          
                        end if;
                    end loop;   

                when write_coeff_abs_level_greater1 =>
                    BinarizeState <= write_coeff_abs_level_greater2;
                    write_coeff_abs_level_greater_1_loop : for index in 0 to 15 loop
                        if sig_coeff_flag(index) = '1' and ALG1_index < 8 then
							DataOut(OutputWidth-1-ALG1_index) <= coeff_abs_level_greater1_flag(index);
                            ALG1_index := ALG1_index + 1;
                        end if;   
                    end loop; 

                    DataLength <= std_logic_vector(to_unsigned(ALG1_index,OutputWidthLength));

                when write_coeff_abs_level_greater2 =>
                    BinarizeState <= write_coeff_abs_level_remaining;
                    write_coeff_abs_level_greater_2_loop : for index in 0 to 15 loop
                        if sig_coeff_flag(index) = '1' then
                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                            DataOut(OutputWidth-1) <= or_reduce(coeff_abs_level_greater2_flag);
                            exit write_coeff_abs_level_greater_2_loop;
                        end if;   
                    end loop; 
                    


                when write_coeff_abs_level_remaining =>
                    if(ABS_done = '1') then
                        BinarizeState <= write_coeff_sign_flag;
                    end if;
                    
                    --DataOut <= '0' & kTable(k)(5);

                    temploop : for i in 0 to 17 loop
                        if (ABS_level_writeout - conv_integer(kTable(k)(i)) >= 0) then   
                            --xorresult := i;
                            DataOut <= '0' & std_logic_vector(to_unsigned(ABS_level_writeout-3 - conv_integer(kTable(k)(i)),16));
                            --exit temploop;
                        end if;
                    end loop;


                    --DataOut <= '0' & std_logic_vector(to_unsigned(ABS_level_writeout - conv_integer(kTable(k)(xorresult)),16));


                when write_coeff_sign_flag =>
                    BinarizeState <= write_finished;
                    write_coeff_sign_flag_loop : for index in 0 to 15 loop
                        if sig_coeff_flag(index) = '1' then
                            if coeff_sign_flag(index) = '1' then
                                DataOut(OutputWidth-1 - SIGN_index) <= '1';
                            else
                                DataOut(OutputWidth-1 - SIGN_index) <= '0';
                            end if;
                            SIGN_index := SIGN_index + 1;
                        end if;
                        DataLength <= std_logic_vector(to_unsigned(SIGN_index,OutputWidthLength));
                    end loop;
                when others =>
                    BinarizeState <= write_finished;
                    Finished <= '1';
                    DataOut        <= (others => '0');
                    DataLength     <= (others => '0');
            end case;
        end if;
    end process;
end struct;